import numpy as np
from skimage.measure import label
cimport fdip_cpp_extension as cpp
from libc.stdint cimport uint32_t, uint16_t, uint8_t

cdef class FastDiffractionImageProcessing:
    cdef cpp.detectorRawFormat_t _detector_raw_format

    cdef cpp.streakFinder_constantArguments_t _streak_finder__constant_arguments
    cdef bint _streak_finder__constant_arguments__allocated

    cdef cpp.radialBackgroundSubtraction_constantArguments_t _radial_rank_filter__constant_arguments
    cdef bint _radial_background_subtraction__constant_arguments__allocated

    cdef cpp.ringFinder_constantArguments_t _ring_finder__constant_arguments
    cdef bint _ring_finder__constant_arguments__allocated


    cdef float[:,::1] _x_map, _y_map, _r_map
    cdef cpp.detectorGeometryMatrix_pythonWrapper_t _detector_geometry_matrix

    cdef uint8_t[:,::1] _global_mask
    cdef uint8_t[:,::1] _peak_finding_mask
    cdef float[:,::1] _data

    cdef cpp.peakList_t _peak_list
    cdef bint _peak_list__allocated

    def __cinit__(self,
        uint16_t asic_nx, uint16_t asic_ny, uint8_t nasics_x, uint8_t nasics_y,
        float[:,::1] x_map,
        float[:,::1] y_map,
        float[:,::1] r_map,
        uint8_t[:,::1] global_mask,
        uint8_t[:,::1] peak_finding_mask,
        uint8_t maskGood,
        uint8_t maskBad
        ):
        self._set_detector_raw_format(asic_nx, asic_ny, nasics_x, nasics_y)
        
        cpp.convertMask(&global_mask[0,0], maskGood, maskBad, self._detector_raw_format);
        cpp.convertMask(&peak_finding_mask[0,0], maskGood, maskBad, self._detector_raw_format);
        
        self._set_detector_geometry_matrix(x_map, y_map, r_map)
        self.set_global_mask(global_mask)
        self.set_peak_finding_mask(peak_finding_mask)
        self._data = np.zeros(shape=(global_mask.shape[0], global_mask.shape[1]) , dtype='float32', order='C') #shape=mask.shape[0:2] does not work

        self._radial_background_subtraction__constant_arguments__allocated = False
        self._streak_finder__constant_arguments__allocated = False

    def _set_detector_raw_format(self, uint16_t asic_nx, uint16_t asic_ny, uint8_t nasics_x, uint8_t nasics_y):
        self._detector_raw_format.asic_nx = asic_nx
        self._detector_raw_format.asic_ny = asic_ny
        self._detector_raw_format.nasics_x = nasics_x
        self._detector_raw_format.nasics_y = nasics_y
        self._detector_raw_format.pix_nx = asic_nx*nasics_x
        self._detector_raw_format.pix_ny = asic_ny*nasics_y
        self._detector_raw_format.pix_nn = asic_nx*nasics_x * asic_ny*nasics_y

    def _set_detector_geometry_matrix(self,
                                      float[:,::1] x_map,
                                      float[:,::1] y_map,
                                      float[:,::1] r_map
                                      ):
        self._x_map = x_map.copy()
        self._y_map = y_map.copy()
        self._r_map = r_map.copy()

        self._detector_geometry_matrix.detectorGeometryMatrix_x = &self._x_map[0,0]
        self._detector_geometry_matrix.detectorGeometryMatrix_y = &self._y_map[0,0]

    def set_data(self, float[:,::1] data):
        cpp.mergeMaskAndDataIntoDataCopy(&data[0,0], &self._data[0,0], &self._global_mask[0,0], self._detector_raw_format)

    def precompute_streak_finder(self,
                                 uint8_t filter_length,
                                 uint8_t min_filter_length,
                                 float filter_step,
                                 float min_snr,
                                 uint8_t streak_elongation_min_steps_count,
                                 float streak_elongation_radius_factor,
                                 uint8_t streak_pixel_mask_radius,
                                 uint16_t[::1] streak_start_pixel_candidates_x = None,
                                 uint16_t[::1] streak_start_pixel_candidates_y = None,
                                 streak_start_pixel_candidates__mask = None, # ndarray, replaces streak_start_pixel_candidates_x/y
                                 uint16_t[::1] background_estimation_regions__upper_left_corner_x = None,
                                 uint16_t[::1] background_estimation_regions__upper_left_corner_y = None,
                                 uint16_t[::1] background_estimation_regions__lower_right_corner_x = None,
                                 uint16_t[::1] background_estimation_regions__lower_right_corner_y = None,
                                 background_estimation_regions__mask = None # ndarray, replaces background_estimation_regions__[corners]
                                 ):
        print('start precomputing streakFinder')

        if streak_start_pixel_candidates__mask is not None:
            (streakStartPixelCandidates_y_np, streakStartPixelCandidates_x_np) = np.where(streak_start_pixel_candidates__mask == 1)
            streakStartPixelCandidates_x_np = np.ascontiguousarray(streakStartPixelCandidates_x_np, dtype='uint16')
            streakStartPixelCandidates_y_np = np.ascontiguousarray(streakStartPixelCandidates_y_np, dtype='uint16')
            streak_start_pixel_candidates_x = streakStartPixelCandidates_x_np
            streak_start_pixel_candidates_y = streakStartPixelCandidates_y_np

        cdef int i
        if background_estimation_regions__mask is not None:
            labeld_background_regions, labels_count = label(background_estimation_regions__mask, return_num=True)

            background_estimation_regions__upper_left_corner_x = np.zeros(labels_count, dtype=np.uint16)
            background_estimation_regions__upper_left_corner_y = np.zeros(labels_count, dtype=np.uint16)
            background_estimation_regions__lower_right_corner_x = np.zeros(labels_count, dtype=np.uint16)
            background_estimation_regions__lower_right_corner_y = np.zeros(labels_count, dtype=np.uint16)
            for i in range(labels_count):
                y,x = np.nonzero(labeld_background_regions == i+1)
                background_estimation_regions__upper_left_corner_x[i] = min(x)
                background_estimation_regions__upper_left_corner_y[i] = min(y)
                background_estimation_regions__lower_right_corner_x[i] = max(x)
                background_estimation_regions__lower_right_corner_y[i] = max(y)

        if self._streak_finder__constant_arguments__allocated:
            cpp.freePrecomputedStreakFinderConstantArguments(self._streak_finder__constant_arguments)

        cdef cpp.streakFinder_accuracyConstants_pythonWrapper_t ac
        ac.filterLength = filter_length
        ac.minFilterLength = min_filter_length
        ac.filterStep = filter_step

        ac.minSNR = min_snr
        ac.streakElongationMinStepsCount = streak_elongation_min_steps_count
        ac.streakElongationRadiusFactor = streak_elongation_radius_factor
        ac.streakPixelMaskRadius = streak_pixel_mask_radius

        ac.streakStartPixelCandidates_x = &streak_start_pixel_candidates_x[0]
        ac.streakStartPixelCandidates_y = &streak_start_pixel_candidates_y[0]
        ac.streakStartPixelCandidatesCount = streak_start_pixel_candidates_y.size

        ac.backgroundEstimationRegions_upperLeftCorner_x = &background_estimation_regions__upper_left_corner_x[0]
        ac.backgroundEstimationRegions_upperLeftCorner_y = &background_estimation_regions__upper_left_corner_y[0]
        ac.backgroundEstimationRegions_lowerRightCorner_x = &background_estimation_regions__lower_right_corner_x[0]
        ac.backgroundEstimationRegions_lowerRightCorner_y = &background_estimation_regions__lower_right_corner_y[0]
        ac.backgroundEstimationRegionsCount = background_estimation_regions__upper_left_corner_x.size


        self._streak_finder__constant_arguments = cpp.precomputeStreakFinderConstantArguments(
            ac,
            self._detector_raw_format,
            self._detector_geometry_matrix,
            &self._global_mask[0,0])

        self._streak_finder__constant_arguments__allocated = True

        print('finished precomputing streakFinder\n')

    def precompute_background_subtraction(self,
        float[:,::1] r_map,
        uint32_t min_values_per_bin,
        uint32_t min_bin_width,
        uint32_t max_considered_values_per_bin,
        uint8_t[::1] detectors_to_consider_subscripts_x,
        uint8_t[::1] detectors_to_consider_subscripts_y,
        uint8_t[::1] detectors_to_correct_subscripts_x,
        uint8_t[::1] detectors_to_correct_subscripts_y,
        float rank,
        float rankSmoothing
        ):
        print('start precomputing background subtraction')

        if(self._radial_background_subtraction__constant_arguments__allocated):
            cpp.freePrecomputeRadialBackgroundSubtractionConstants(self._radial_rank_filter__constant_arguments)

        if detectors_to_consider_subscripts_x.size == 0:
            (detectorsToConsiderSubscripts_x_np, detectorsToConsiderSubscripts_y_np) = \
                np.meshgrid(np.arange(self._detector_raw_format.nasics_x), np.arange(self._detector_raw_format.nasics_y))
            detectors_to_consider_subscripts_x = detectorsToConsiderSubscripts_x_np.ravel().astype(np.uint8)
            detectors_to_consider_subscripts_y = detectorsToConsiderSubscripts_y_np.ravel().astype(np.uint8)
        if detectors_to_correct_subscripts_x.size == 0:
            (detectorsToCorrectSubscripts_x_np, detectorsToCorrectSubscripts_y_np) = \
                np.meshgrid(np.arange(self._detector_raw_format.nasics_x), np.arange(self._detector_raw_format.nasics_y))
            detectors_to_correct_subscripts_x = detectorsToCorrectSubscripts_x_np.ravel().astype(np.uint8)
            detectors_to_correct_subscripts_y = detectorsToCorrectSubscripts_y_np.ravel().astype(np.uint8)

        cdef cpp.radialBackgroundSubtraction_accuracyConstants_pythonWrapper_t ac
        ac.minValuesPerBin = min_values_per_bin
        ac.minBinWidth = min_bin_width
        ac.maxConsideredValuesPerBin = max_considered_values_per_bin
        ac.detectorsToConsiderSubscripts_x = &detectors_to_consider_subscripts_x[0]
        ac.detectorsToConsiderSubscripts_y = &detectors_to_consider_subscripts_y[0]
        ac.detectorsToConsiderCount = detectors_to_consider_subscripts_x.size
        ac.detectorsToCorrectSubscripts_x = &detectors_to_correct_subscripts_x[0]
        ac.detectorsToCorrectSubscripts_y = &detectors_to_correct_subscripts_y[0]
        ac.detectorsToCorrectCount = detectors_to_correct_subscripts_x.size
        ac.rank = rank
        ac.rankSmoothing = rankSmoothing

        self._radial_rank_filter__constant_arguments = cpp.precomputeRadialBackgroundSubtractionConstantArguments(
            &self._global_mask[0,0], &r_map[0,0], self._detector_raw_format, ac, self._detector_geometry_matrix)

        self._radial_background_subtraction__constant_arguments__allocated = True
        print('finished precomputing radial background subtraction\n')

    def precompute_ring_finder(self,
                              float relaltiveRingDiff,
                              uint16_t smoothWindowRadius,
                              uint16_t ringDilationRadius,
                              uint16_t minRingRadius,
                              uint16_t maxRingRadius
                              ):
        print('start precomputing ring finder')

        cdef cpp.ringFinder_accuracyConstants_t ac
        ac.relaltiveRingDiff = relaltiveRingDiff
        ac.smoothWindowRadius = smoothWindowRadius
        ac.ringDilationRadius = ringDilationRadius
        ac.minRingRadius = minRingRadius
        ac.maxRingRadius = maxRingRadius

        if(self._ring_finder__constant_arguments__allocated):
            cpp.freePrecomputedRingFinderConstantArguments(self._ring_finder__constant_arguments)

        self._ring_finder__constant_arguments = cpp.precomputeRingFinderConstantArguments(&self._r_map[0,0], ac, self._detector_raw_format)

        self._ring_finder__constant_arguments__allocated = True
        print('finished precomputing ring finder\n')

    def apply_ring_finder(self):
        if not self._ring_finder__constant_arguments__allocated:
            print("precompute ringFinder before using it")
            return
        cpp.applyRingFinder(&self._data[0,0], self._ring_finder__constant_arguments)

    def precompute_peak_finder9(self, uint32_t max_peak_count):
        if(self._peak_list__allocated):
            cpp.freePeakList(self._peak_list)

        cdef int memoryAllocationErrorFlag
        memoryAllocationErrorFlag = cpp.allocatePeakList(&self._peak_list, max_peak_count)

        self._peak_list__allocated = True

        if(memoryAllocationErrorFlag):
            raise MemoryError()

    def apply_peak_finder9(self,
        float min_snr_biggest_pixel,              # small snr leads to a slow algorithm
        float min_snr_peak_pixel,                 # should be smaller or equal to minSNR_biggestPixel
        float min_snr_whole_peak,                 # should be bigger or equal to minSNR_biggestPixel
        float minimum_sigma,                         # to not find false peaks in very dark noise free regions
        float minimum_peak_oversize_over_neighbours,    # for faster processing
        uint8_t window_radius
        ):

        if not self._peak_list__allocated:
            print("precompute peakFinder9 before using it")
            return

        cpp.mergeMaskIntoData(&self._data[0,0], &self._peak_finding_mask[0,0], self._detector_raw_format)

        cdef cpp.peakFinder9_accuracyConstants_t ac
        ac.minSNR_biggestPixel = min_snr_biggest_pixel
        ac.minSNR_peakPixel = min_snr_peak_pixel
        ac.minSNR_wholePeak = min_snr_whole_peak
        ac.minimumSigma = minimum_sigma
        ac.minimumPeakOversizeOverNeighbours = minimum_peak_oversize_over_neighbours
        ac.windowRadius = window_radius

        self._peak_list.peakCount = 0
        cdef uint32_t peakCount = cpp.peakFinder9(&self._data[0,0], ac, self._detector_raw_format, self._peak_list)

        max_intensity = np.zeros(peakCount, dtype=float)      # Maximum intensity in peak
        total_intensity = np.zeros(peakCount, dtype=float)    # Integrated intensity in peak
        sigma_background = np.zeros(peakCount, dtype=float)   # Standard deviation of the background
        snr = np.zeros(peakCount, dtype=float)               # Signal-to-noise ratio of peak
        pixel_count = np.zeros(peakCount, dtype=float)        # Number of pixels in peak
        center_of_mass__raw_x = np.zeros(peakCount, dtype=float) # peak center of mass x (in raw layout)
        center_of_mass__raw_y = np.zeros(peakCount, dtype=float) # peak center of mass y (in raw layout)

        for i in range(0, peakCount):
            max_intensity[i] = self._peak_list.maxIntensity[i]
            total_intensity[i] = self._peak_list.totalIntensity[i]
            sigma_background[i] = self._peak_list.sigmaBackground[i]
            snr[i] = self._peak_list.snr[i]
            pixel_count[i] = self._peak_list.pixelCount[i]
            center_of_mass__raw_x[i] = self._peak_list.centerOfMass_rawX[i]
            center_of_mass__raw_y[i] = self._peak_list.centerOfMass_rawY[i]

        return (max_intensity, total_intensity, sigma_background, snr, pixel_count, center_of_mass__raw_x, center_of_mass__raw_y)

    def get_data(self):
        return np.asarray(self._data)

    def set_global_mask(self, uint8_t[:,::1] mask):
        self._global_mask = mask.copy()
        # create sparse mask if mask is sparse

    def set_peak_finding_mask(self, uint8_t[:,::1] mask):
        self._peak_finding_mask = mask.copy()
        # create sparse mask if mask is sparse. Need only to take diff to _global_mask

    def get_computed_mask(self):
        cdef uint8_t[:,::1] mask = np.zeros(shape=(self._global_mask.shape[0], self._global_mask.shape[1]) , dtype='uint8', order='C') #shape=_global_mask.shape[0:2] does not work
        cpp.getMaskFromMergedMaskInData(&self._data[0,0], &mask[0,0], self._detector_raw_format)
        return np.asarray(mask)

    def apply_streak_finder(self):
        if not self._streak_finder__constant_arguments__allocated:
            print("precompute streakFinder before using it")
            return
        cpp.streakFinder(&self._data[0,0], self._streak_finder__constant_arguments)

    def apply_radial_rank_filter(self):
        if not self._radial_background_subtraction__constant_arguments__allocated:
            print("precompute radial rank filter before using it")
            return
        cpp.applyRadialBackgroundSubtraction(&self._data[0,0], self._radial_rank_filter__constant_arguments)

    def __dealloc__(self):
        if(self._streak_finder__constant_arguments__allocated):
            cpp.freePrecomputedStreakFinderConstantArguments(self._streak_finder__constant_arguments)

        if(self._radial_background_subtraction__constant_arguments__allocated):
            cpp.freePrecomputeRadialBackgroundSubtractionConstants(self._radial_rank_filter__constant_arguments)

        if(self._ring_finder__constant_arguments__allocated):
            cpp.freePrecomputedRingFinderConstantArguments(self._ring_finder__constant_arguments)

        if(self._peak_list__allocated):
            cpp.freePeakList(self._peak_list)
