#include "fastDiffractionImageProcessing/ringFinder.h"

#include <cfloat>
#include <set>

using namespace std;
using namespace Eigen;

static void medianFilter(const float *data, int valuesCount, int windowRadius, float *medianFiltered);
static inline float getMultisetMedian(const multiset< float >& mymultiset);

void precomputeRingFinderConstants(ringFinder_precomputedConstants_t& precomputedConstants, const float* detectorGeometryRadiusMatrix_linear,
        const ringFinder_accuracyConstants_t& accuracyConstants, const detectorRawFormat_t& detectorRawFormat)
{

    precomputedConstants.binIndices.resize(detectorRawFormat.pix_nn);
    for (uint32_t i = 0; i < detectorRawFormat.pix_nn; ++i) {
        precomputedConstants.binIndices[i] = (uint16_t) floor(detectorGeometryRadiusMatrix_linear[i]);
    }

    int binCount = *max_element(precomputedConstants.binIndices.begin(), precomputedConstants.binIndices.end()) + 1;
    precomputedConstants.dataCountInBins = ArrayXf::Zero(binCount);
    for (uint32_t i = 0; i < detectorRawFormat.pix_nn; ++i) {
        precomputedConstants.dataCountInBins[precomputedConstants.binIndices[i]]++;
    }

    precomputedConstants.bins.resize(binCount);
    precomputedConstants.medianFilteredBins.resize(binCount);
    precomputedConstants.dataIndicesInBins.clear();
    precomputedConstants.dataIndicesInBins.resize(binCount);
    for (int i = 0; i < binCount; ++i) {
        precomputedConstants.dataIndicesInBins[i].reserve(precomputedConstants.dataCountInBins[i]);
    }
    for (uint32_t i = 0; i < detectorRawFormat.pix_nn; ++i) {
        precomputedConstants.dataIndicesInBins[precomputedConstants.binIndices[i]].push_back(i);
    }

    precomputedConstants.detectorRawFormat = detectorRawFormat;
    precomputedConstants.accuracyConstants = accuracyConstants;
}

void ringFinder(float* data, ringFinder_precomputedConstants_t& precomputedConstants)
{
    precomputedConstants.dataCountInBins.setZero();
    precomputedConstants.bins.setZero();
    for (uint32_t i = 0; i < precomputedConstants.detectorRawFormat.pix_nn; ++i) {
        if (data[i] != -FLT_MAX) {
            precomputedConstants.dataCountInBins[precomputedConstants.binIndices[i]]++;
            precomputedConstants.bins[precomputedConstants.binIndices[i]] += data[i];
        }
    }

    precomputedConstants.bins /= precomputedConstants.dataCountInBins;

    medianFilter(precomputedConstants.bins.data(), precomputedConstants.bins.size(),
            precomputedConstants.accuracyConstants.smoothWindowRadius, precomputedConstants.medianFilteredBins.data());

    ArrayXf relativeDifference = abs((precomputedConstants.bins - precomputedConstants.medianFilteredBins) / precomputedConstants.bins);

    ArrayXf relativeDifferenceMedianFiltered(relativeDifference.size());
    int windowRadius = 3;
    medianFilter(relativeDifference.data(), relativeDifference.size(), windowRadius, relativeDifferenceMedianFiltered.data());

    int dilationRadius = precomputedConstants.accuracyConstants.ringDilationRadius;
    set< int > binsToMask;
    int minRadius = precomputedConstants.accuracyConstants.minRingRadius;
    int maxRadius = precomputedConstants.accuracyConstants.maxRingRadius;
    for (int i = max(0, minRadius); i < min((int) relativeDifferenceMedianFiltered.size(), maxRadius); ++i) {
        if (relativeDifferenceMedianFiltered[i] >= precomputedConstants.accuracyConstants.relaltiveRingDiff) {
            for (int j = max(0, i - dilationRadius); j <= min(int(relativeDifferenceMedianFiltered.size() - 1), i + dilationRadius); ++j) {
                binsToMask.insert(j);
            }
        }
    }

    for (const auto& binToMask : binsToMask) {
        for (const auto& maskIndex : precomputedConstants.dataIndicesInBins[binToMask]) {
            data[maskIndex] = -FLT_MAX;
        }
    }
}

static float getMultisetMedian(const multiset< float >& mymultiset)
{
    auto it = mymultiset.cbegin();
    for (size_t i = 0; i < mymultiset.size() / 2; i++) {
        it++;
    }
    return *it;
}

static void medianFilter(const float *data, int valuesCount, int windowRadius, float *medianFiltered)
{
    int windowSize = windowRadius * 2 + 1;

    if (windowSize > valuesCount) {
        windowSize = windowRadius * 2 + 1;
        windowRadius = windowSize / 2;
    }

    multiset< float > m;
    m.insert(data, data + windowRadius);

    for (int i = 0; i < windowRadius; ++i) {
        m.insert(data[i + windowRadius]);
        medianFiltered[i] = getMultisetMedian(m);
    }

    for (int i = windowRadius; i < valuesCount - windowRadius - 1; ++i) {
        m.insert(data[i + windowRadius]);
        medianFiltered[i] = getMultisetMedian(m);
        m.erase(m.find(data[i - windowRadius]));
    }

    for (int i = valuesCount - windowRadius - 1; i < valuesCount; ++i) {
        medianFiltered[i] = getMultisetMedian(m);
        m.erase(m.find(data[i - windowRadius]));
    }
}
